#include "render.h"
#include "anorm.h"

#define MAX_POLYGONS 4800 //Attempt to match SGL specification [in my case, sending a polygon is far more expensive than transforming a vert]
#define MAX_VERTS 4000 //Attempt to match SGL specification 
#define MAX_MSH2_SENT_POLYS (600)
#define UNCACHE (0x20000000)

volatile int * DVSR = (volatile int*)0xFFFFFF00;
volatile int * DVDNTH = (volatile int*)0xFFFFFF10;
volatile int * DVDNTL = (volatile int*)0xFFFFFF14;

SPRITE * localSprBuf = (SPRITE *)0x060D5B60;

//Write-only register! of CRAMoffset[1], bits 6,5,4 control SPR layer offset.
unsigned short * vdp2_CRAMoffset = (unsigned short *)(0x1800E4 + 0x25E00000);
//

vertex_t * ssh2VertArea;
vertex_t * msh2VertArea;
animationControl * AnimArea;

int dummy[4];
int * ssh2SentPolys;
int * msh2SentPolys;
int * transVerts;
int * transPolys;
int anims;

const	FIXED	nearP	= 5<<16; // Z DISPLAY LEVEL [Actually a bit more complicated than that, but it works]
const	FIXED	farP	= 800<<16; // RENDER DIST

FIXED outTime = 0;
FIXED lasttime = 0;
Sint32 framerate;
Sint8 SynchConst;

void	init_render_area(void){
	
	ssh2VertArea = (vertex_t *)(unsigned int)jo_malloc(1024 * sizeof(vertex_t));
	msh2VertArea = (vertex_t *)(unsigned int)jo_malloc(1024 * sizeof(vertex_t));
	AnimArea = (animationControl *)(unsigned int)jo_malloc(16 * sizeof(animationControl));
	localSprBuf = (SPRITE *)((unsigned int)jo_malloc( (2 * sizeof(SPRITE)) ));
	ssh2SentPolys = (int *)(((unsigned int)&dummy[0])|UNCACHE);
	msh2SentPolys = (int *)(((unsigned int)&dummy[1])|UNCACHE);
	transVerts = (int *)(((unsigned int)&dummy[2])|UNCACHE);
	transPolys = (int *)(((unsigned int)&dummy[3])|UNCACHE);
}

void	frame_render_prep(void)
{
		ssh2SentPolys[0] = 0;
		msh2SentPolys[0] = 0;
		transVerts[0] = 0;
		transPolys[0] = 0;
		anims = 0;
		slSetSprite(&localSprBuf[0], (50<<16));
}

//borrowed/given by XL2 -- Frame limiter to 30 FPS. EXTREMELY USEFUL.
void	update_gamespeed(void)
{
//	static int curtime = 0;
    outTime = jo_get_ticks(); //outTime is global variable

	int frmrt = (outTime-lasttime);
	
	static Uint8 lastTimes[100];
	static Uint8 time_selector = 0;
	static Uint8 worst_frame = 0;
	static Uint16 bad_times = 0;
	lastTimes[time_selector] = frmrt;
	time_selector++;
	if(time_selector > 100){
		time_selector = 0;
		worst_frame = 0;
		for(Uint8 c =0; c < 100; c++) lastTimes[c] = 0;
	}
	for(Uint8 n = 0; n < 100; n++){
		worst_frame = (lastTimes[n] > worst_frame) ? lastTimes[n] : worst_frame;
	}
	if(frmrt > 40) bad_times++;
	
    framerate = (frmrt)>>4;
    lasttime = outTime;
	
    if (framerate <= 0) framerate=1;
    else if (framerate > 5) framerate=5;
	jo_printf(0, 2, "x(%i)x", frmrt);
	jo_printf(8, 3, "1/100 Low:(%i)", worst_frame);
	jo_printf(8, 2, "Bad Frames:(%i)", bad_times);
	slPrintFX(outTime, slLocate(26, 2));
}


FIXED		fxm(FIXED d1, FIXED d2) //Fixed Point Multiplication
{
	register FIXED rtval;
	asm(
	"dmuls.l %[d1],%[d2];"
	"sts MACH,r1;"
	"sts MACL,%[out];"
	"xtrct r1,%[out];"
    :    [out] "=r" (rtval)       		 //OUT
    :    [d1] "r" (d1), [d2] "r" (d2)    //IN
	:		"r1"						//CLOBBERS
	);
	return rtval;
}


FIXED	trans_pt_by_component(POINT * ptx, FIXED * normal)
{
	register FIXED transPt;
	asm(
		"clrmac;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"sts MACH,r0;"
		"sts MACL,%[ox];"
		"xtrct r0,%[ox];"
		"mov.l @%[nmptr],r0;"
		"add r0,%[ox];"
		: 	[ox] "=r" (transPt)											//OUT
		:	[ptptr] "r" (ptx) ,	[nmptr] "r" (normal)					//IN
		:	"r0"														//CLOBBERS
	);
	return transPt;
}

FIXED	norm_component_trans(VECTOR * norm, FIXED * mtxNorm) //Same as above, only position is not added.
{
	register FIXED transY;
	asm(
		"clrmac;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"mac.l @%[ptptr]+,@%[nmptr]+;"
		"sts MACH,r0;"
		"sts MACL,%[ox];"
		"xtrct r0,%[ox];"
		: 	[ox] "=r" (transY)											//OUT
		:	[ptptr] "r" (norm) , [nmptr] "r" (mtxNorm)					//IN
		:	"r0"														//CLOBBERS
	);
	return transY;
}

void	animation_interpolator(short * curKeyFrame, short * nextKeyFrame, unsigned int frDelta, Sint32 * pntbl)
{
	//Interpolates single-component of vertices.
	//The pntbl pointer is to be its own unique pointer. Do not use or modify an external pointer. Do not use an array indice,
	//or address of an array member. Initialize as pdata->pntbl[0].
	asm(
	"mov.w @%[curKeyPtr]+,r4;" //This instruction will increment curKeyFrame [these contain compressed, but complete, vertice data]
	"mov.w @%[nxtKeyPtr]+,r5;" //This instruction will increment nextKeyFrame [by 'complete' i mean they are not delta-data themselves]
	"sub r4,r5;" //Component data delta between current and next keyframe
	"muls.w r5, %[frDelta];" //Multiplied by absolute delta (16 x 16 -> 32) / Expresses our distance from one keyframe to the other, in per-frame terms.
	"sts MACL,r5;" //The frDelta is not vertice component data!
	"shlr2 r5;" //We want a proportional delta between e.g. cur[X] and nxt[X]. So we shift it down a bit, which is a division.
	"shlr2 r5;" //My theory is, if this is 8.8 data, the frame delta data isn't all we have nor is it all we want. So we bind it to 4.8 instead.
	"add r4,r5;" //Add this to the current key frame data (if frame delta is 0, no change!)
	"shll8 r5;" //MSB of prior ops is 15 ; so 8.8 data. This moves it up to be bound on the right bit for 16.16 data.
	"mov.l r5, @%[pntbl];" //This instruction moves the result out to memory
	"mov.l @%[pntbl]+,r3;" //Dummy instruction to increment pointer
    :							//OUT
    :    [pntbl] "r" (pntbl), [curKeyPtr] "r" (curKeyFrame), [nxtKeyPtr] "r" (nextKeyFrame), [frDelta] "r" (frDelta)	//IN
	:		"r5" , "r4", "r3"	//CLOBBERS
	);
	
}

//Set data in registers for division unit.
void		SetFixDiv(FIXED dividend, FIXED divisor) //Defined as "dividend / divisor", for fixed points, using division unit
{

/*
SH7604 Manual Information:

The 64-bit dividend is set in dividend registers H and L (DVDNTH and DVDNTL).
First set the value in DVDNTH. When a value is written to DVDNTL, the 64-bit ÷ 32-bit operation begins.
After the operation, the 32-bit remainder is written to DVDNTH and the 32-bit quotient is written to DVDNTL.

[ME:]These registers can only be accessed via pointers. . . because our compiler is not aware of them.
*/	

DVSR[0] = divisor;
DVDNTH[0] = (dividend>>16);
DVDNTL[0] = (dividend<<16);
	
}

static inline void	ssh2SetCommand(FIXED * p1, FIXED * p2, FIXED * p3, FIXED * p4, Uint16 cmdctrl,
                                Uint16 cmdpmod, Uint16 cmdsrca, Uint16 cmdcolr,
                                Uint16 cmdsize, Uint16 cmdgrda, FIXED drawPrty) {

	SPRITE_T * user_sprite = (SPRITE_T *)&SpriteBuf[ssh2SentPolys[0] + MAX_MSH2_SENT_POLYS];
	ssh2SentPolys[0]++;

   user_sprite->CTRL = cmdctrl;
   user_sprite->LINK = 0x3000;
   user_sprite->PMOD = cmdpmod;
   user_sprite->SRCA = cmdsrca; //TEXTURE ADDRESS IN VDP1 VRAM
   user_sprite->COLR = cmdcolr; //COLOR BANK CODE IN COLOR BANK MODES, DRAW COLOR IN UNTEXTURED MODES, LUT ADDRESS IN CL16LK, IGNORED RGB
   user_sprite->SIZE = cmdsize; //VALID FOR TEXTURE DRAW COMMANDS

   user_sprite->XA=p1[X];
   user_sprite->YA=p1[Y];
   user_sprite->XB=p2[X];
   user_sprite->YB=p2[Y];
   user_sprite->XC=p3[X];
   user_sprite->YC=p3[Y];
   user_sprite->XD=p4[X];
   user_sprite->YD=p4[Y];
   user_sprite->DMMY=(drawPrty>>16);

   /**Important : Only the slave CPU is allowed to read/write the z sort buffer**/
   //IMPORTANT: We have to use the "far" screen. This is the "128". Why? Someone got angry when I asked why...
   //CRITICAL!! You MUST change sl_def to include the SPRITE_T structure. It is the same as SPRITE except with a Uint32 pointer, NEXT.
   //You must also change SpriteBuf and SpriteBuf2 to be of SPRITE_T * type.
    Uint32 ** ptr = (Uint32**)(Zbuffer + (128 + ((user_sprite->DMMY>>8)))*4  );
    user_sprite->NEXT=*ptr;
    *ptr=(void*)user_sprite;
}

static inline void	msh2SetCommand(FIXED * p1, FIXED * p2, FIXED * p3, FIXED * p4, Uint16 cmdctrl,
                                Uint16 cmdpmod, Uint16 cmdsrca, Uint16 cmdcolr,
                                Uint16 cmdsize, Uint16 cmdgrda, FIXED drawPrty) {

	SPRITE_T * user_sprite = (SPRITE_T *)&SpriteBuf[msh2SentPolys[0]];
	msh2SentPolys[0]++;

   user_sprite->CTRL = cmdctrl;
   user_sprite->LINK = 0x3000;
   user_sprite->PMOD = cmdpmod;
   user_sprite->SRCA = cmdsrca; //TEXTURE ADDRESS IN VDP1 VRAM
   user_sprite->COLR = cmdcolr; //COLOR BANK CODE IN COLOR BANK MODES, DRAW COLOR IN UNTEXTURED MODES, LUT ADDRESS IN CL16LK, IGNORED RGB
   user_sprite->SIZE = cmdsize; //VALID FOR TEXTURE DRAW COMMANDS

   user_sprite->XA=p1[X];
   user_sprite->YA=p1[Y];
   user_sprite->XB=p2[X];
   user_sprite->YB=p2[Y];
   user_sprite->XC=p3[X];
   user_sprite->YC=p3[Y];
   user_sprite->XD=p4[X];
   user_sprite->YD=p4[Y];
   user_sprite->DMMY=(drawPrty>>16);
   

}

void	sort_master_polys(void)
{
	SPRITE_T * user_sprite;
	
	for(int i = 0; i < msh2SentPolys[0]; i++)
	{
	user_sprite = (SPRITE_T *)&SpriteBuf[i];
		
   /**Important : Only the slave CPU is allowed to read/write the z sort buffer**/
   //IMPORTANT: We have to use the "far" screen. This is the "128". Why? Someone got angry when I asked why...
   //CRITICAL!! You MUST change sl_def to include the SPRITE_T structure. It is the same as SPRITE except with a Uint32 pointer, NEXT.
   //You must also change SpriteBuf and SpriteBuf2 to be of SPRITE_T * type.
    Uint32 ** ptr = (Uint32**)(Zbuffer + (128 + ((user_sprite->DMMY>>8)))*4  );
    user_sprite->NEXT=*ptr;
    *ptr=(void*)user_sprite;
	}
}

void ssh2DrawModel(entity_t * ent, FIXED * lightSrc)
{
	//Recommended, for performance, that large entities be placed in HWRAM.
    static MATRIX newMtx;
    slGetMatrix(newMtx);

	static FIXED m0x[4];
	static FIXED m1y[4];
	static FIXED m2z[4];
	
	m0x[0] = newMtx[X][X];
	m0x[1] = newMtx[Y][X];
	m0x[2] = newMtx[Z][X];
	m0x[3] = newMtx[3][X];
	
	m1y[0] = newMtx[X][Y];
	m1y[1] = newMtx[Y][Y];
	m1y[2] = newMtx[Z][Y];
	m1y[3] = newMtx[3][Y];
	
	m2z[0] = newMtx[X][Z];
	m2z[1] = newMtx[Y][Z];
	m2z[2] = newMtx[Z][Z];
	m2z[3] = newMtx[3][Z];

    XPDATA * model = ent->pol[0];
    Uint32	cnt;
	Uint32	nbPt	= model->nbPoint;
	Uint32	nbPol	= model->nbPolygon;
    if ( (transVerts[0]+nbPt) >= MAX_VERTS) return;
    if ( (transPolys[0]+nbPol) >= MAX_POLYGONS) return;
   // Uint32 PreClippingflag;
	
	FIXED transNormY;
	short colorBank;

	unsigned short comm_p_mode = 5328; //Should add color mode table? Color mode is applied here.
	unsigned short ctrl = 2;
 
    for (cnt=0; cnt<nbPt; cnt++)
    {
        /**calculate z**/
        ssh2VertArea[cnt].pnt[Z] = trans_pt_by_component(&model->pntbl[cnt], m2z);
		
		ssh2VertArea[cnt].pnt[Z] = (ssh2VertArea[cnt].pnt[Z] > nearP) ? ssh2VertArea[cnt].pnt[Z] : nearP;

 
         /**Starts the division**/
        SetFixDiv(MsScreenDist, ssh2VertArea[cnt].pnt[Z]);

        /**Calculates X and Y while waiting for screenDist/z **/
        ssh2VertArea[cnt].pnt[Y] = trans_pt_by_component(&model->pntbl[cnt], m1y);
		ssh2VertArea[cnt].clipFlag = (ssh2VertArea[cnt].pnt[Z] > nearP) ? 0 : 16; //I have to put this here because my ASM is sloppy
        ssh2VertArea[cnt].pnt[X] = trans_pt_by_component(&model->pntbl[cnt], m0x);
		
        /** Retrieves the result of the division **/
		ssh2VertArea[cnt].inverseZ = *DVDNTL;

        /**Transform X and Y to screen space**/
        ssh2VertArea[cnt].pnt[X] = fxm(ssh2VertArea[cnt].pnt[X], ssh2VertArea[cnt].inverseZ)>>16;
        ssh2VertArea[cnt].pnt[Y] = fxm(ssh2VertArea[cnt].pnt[Y], ssh2VertArea[cnt].inverseZ)>>16;
 
        /**Clipping flags for fast culling/preclipping disable/enable and portal clipping **/
		ssh2VertArea[cnt].clipFlag |= (JO_ABS(ssh2VertArea[cnt].pnt[X]) > JO_TV_WIDTH_2) ? 1 : 0; //Simplified to increase CPU performance
		ssh2VertArea[cnt].clipFlag |= (JO_ABS(ssh2VertArea[cnt].pnt[Y]) > JO_TV_HEIGHT_2) ? 1 : 0; 
    }

    transVerts[0]+=nbPt;

    /**POLYGON PROCESSING**/ 
    for (cnt=0; cnt<nbPol; cnt++)
    {
		//Components of screen-space cross-product used for backface culling.
		//Vertice order hint:
		// 0 - 1
		// 3 - 2
		//A cross-product can tell us if it's facing the screen. If it is not, we do not want it.
		register int cross0 = (ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[X] - ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[X])
							* (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Y] - ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Y]);
		register int cross1 = (ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[Y] - ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[Y])
							* (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[X] - ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[X]);
		//Sorting target. Uses average of top-left and bottom-right. Adding logic to change sorting per-polygon HAMMERS performance in unacceptable ways.
		register int zDepthTgt = (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Z] + ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Z])>>1;
 
 
		if(cross0 >= cross1 || zDepthTgt <= nearP || zDepthTgt >= farP ||
		ssh2VertArea[model->pltbl[cnt].Vertices[0]].clipFlag != 0 ||
		ssh2VertArea[model->pltbl[cnt].Vertices[2]].clipFlag != 0) continue;
		//Preclipping disable logic is removed. It is not worth the CPU performance cost.
		
		//Transform the polygon's normal by light source vector
		transNormY = norm_component_trans(&model->pltbl[cnt].norm, lightSrc);
		//Use transformed normal as shade determinant
		colorBank = (transNormY < -32768) ? 0 : 1;
		colorBank = (transNormY > 16384) ? 2 : colorBank;
		colorBank = (transNormY > 32768) ? 3 : colorBank; 

	//Issues: Unless I load the data the same way in which XL2 does, I will have to adapt the PDATA ATTBL to these functions.
        ssh2SetCommand(ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt,
                     ctrl, comm_p_mode,
                     (int)__jo_sprite_pic[model->attbl[cnt].texno].data, colorBank<<6, __jo_sprite_def[model->attbl[cnt].texno].size, 0, zDepthTgt);
    }
		transPolys[0]+=nbPol;

}

void msh2DrawModel(entity_t * ent, MATRIX msMatrix, FIXED * lightSrc)
{
	//Recommended, for performance, that large entities be placed in HWRAM.
	/**WARNING: DO NOT USE SGL MATRIX SYSTEM FOR MASTER DRAWING**/
	/*
	SGL matrix pointer is uncached. Matrix operations will be confused between master/slave
	You will have to come up with your own matrix system, it's not too bad
	MSH2 + SSH2 drawing is not the ideal. In other words, using 1 CPU at full tap and then only if you must spill over to the second.
	Diminishing returns might be a better phrase. 
	But there is no system here to do that. Using MSH2 does help; so perhaps throw a known quantity at it, rather than any dynamic entity list.
	*/
	static FIXED m0x[4];
	static FIXED m1y[4];
	static FIXED m2z[4];
	
	m0x[0] = msMatrix[X][X];
	m0x[1] = msMatrix[Y][X];
	m0x[2] = msMatrix[Z][X];
	m0x[3] = msMatrix[3][X];
	
	m1y[0] = msMatrix[X][Y];
	m1y[1] = msMatrix[Y][Y];
	m1y[2] = msMatrix[Z][Y];
	m1y[3] = msMatrix[3][Y];
	
	m2z[0] = msMatrix[X][Z];
	m2z[1] = msMatrix[Y][Z];
	m2z[2] = msMatrix[Z][Z];
	m2z[3] = msMatrix[3][Z];

    XPDATA * model = ent->pol[0];
    Uint32	cnt;
	Uint32	nbPt	= model->nbPoint;
	Uint32	nbPol	= model->nbPolygon;
    if ( (transVerts[0]+nbPt) >= MAX_VERTS) return;
    if ( (transPolys[0]+nbPol) >= MAX_POLYGONS) return;
   // Uint32 PreClippingflag;
	
	FIXED transNormY;
	short colorBank;

	unsigned short comm_p_mode = 5328; //Should add color mode table? Color mode is applied here.
	unsigned short ctrl = 2;
 
    for (cnt=0; cnt<nbPt; cnt++)
    {
        /**calculate z**/
        msh2VertArea[cnt].pnt[Z] = trans_pt_by_component(&model->pntbl[cnt], m2z);
		
		msh2VertArea[cnt].pnt[Z] = (msh2VertArea[cnt].pnt[Z] > nearP) ? msh2VertArea[cnt].pnt[Z] : nearP;

 
         /**Starts the division**/
        SetFixDiv(MsScreenDist, msh2VertArea[cnt].pnt[Z]);

        /**Calculates X and Y while waiting for screenDist/z **/
        msh2VertArea[cnt].pnt[Y] = trans_pt_by_component(&model->pntbl[cnt], m1y);
		msh2VertArea[cnt].clipFlag = (msh2VertArea[cnt].pnt[Z] > nearP) ? 0 : 16; //I have to put this here because my ASM is sloppy
        msh2VertArea[cnt].pnt[X] = trans_pt_by_component(&model->pntbl[cnt], m0x);
		
        /** Retrieves the result of the division **/
		msh2VertArea[cnt].inverseZ = *DVDNTL;

        /**Transform X and Y to screen space**/
        msh2VertArea[cnt].pnt[X] = fxm(msh2VertArea[cnt].pnt[X], msh2VertArea[cnt].inverseZ)>>16;
        msh2VertArea[cnt].pnt[Y] = fxm(msh2VertArea[cnt].pnt[Y], msh2VertArea[cnt].inverseZ)>>16;
 
        /**Clipping flags for fast culling/preclipping disable/enable and portal clipping **/
		msh2VertArea[cnt].clipFlag |= (JO_ABS(msh2VertArea[cnt].pnt[X]) > JO_TV_WIDTH_2) ? 1 : 0; //Simplified to increase CPU performance
		msh2VertArea[cnt].clipFlag |= (JO_ABS(msh2VertArea[cnt].pnt[Y]) > JO_TV_HEIGHT_2) ? 1 : 0; 
    }

    transVerts[0]+=nbPt;

    /**POLYGON PROCESSING**/ 
    for (cnt=0; cnt<nbPol; cnt++)
    {
		//Components of screen-space cross-product used for backface culling.
		//Vertice order hint:
		// 0 - 1
		// 3 - 2
		//A cross-product can tell us if it's facing the screen. If it is not, we do not want it.
		register int cross0 = (msh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[X] - msh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[X])
							* (msh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Y] - msh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Y]);
		register int cross1 = (msh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[Y] - msh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[Y])
							* (msh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[X] - msh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[X]);
		//Sorting target. Uses average of top-left and bottom-right. Adding logic to change sorting per-polygon HAMMERS performance in unacceptable ways.
		register int zDepthTgt = (msh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Z] + msh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Z])>>1;
 
		if(cross0 >= cross1 || zDepthTgt <= nearP || zDepthTgt >= farP ||
		msh2VertArea[model->pltbl[cnt].Vertices[0]].clipFlag != 0 ||
		msh2VertArea[model->pltbl[cnt].Vertices[2]].clipFlag != 0 || msh2SentPolys[0] >= MAX_MSH2_SENT_POLYS) continue;
		//Preclipping disable logic is removed. It is not worth the CPU performance cost.
		
		//Transform the polygon's normal by light source vector
		transNormY = norm_component_trans(&model->pltbl[cnt].norm, lightSrc);
		//Use transformed normal as shade determinant
		colorBank = (transNormY < -32768) ? 0 : 1;
		colorBank = (transNormY > 16384) ? 2 : colorBank;
		colorBank = (transNormY > 32768) ? 3 : colorBank; 

	//Issues: Unless I load the data the same way in which XL2 does, I will have to adapt the PDATA ATTBL to these functions.
        msh2SetCommand(msh2VertArea[model->pltbl[cnt].Vertices[0]].pnt,
		msh2VertArea[model->pltbl[cnt].Vertices[1]].pnt,
		msh2VertArea[model->pltbl[cnt].Vertices[2]].pnt,
		msh2VertArea[model->pltbl[cnt].Vertices[3]].pnt,
                     ctrl, comm_p_mode,
                     (int)__jo_sprite_pic[model->attbl[cnt].texno].data, colorBank<<6, __jo_sprite_def[model->attbl[cnt].texno].size, 0, zDepthTgt);
    }
		transPolys[0]+=nbPol;

}

void ssh2DrawAnimation(entity_t * ent, animationControl * animCtrl, FIXED * lightSrc)
{
	//if(ent->file_done != true){return;}
	//WARNING:
	//Once an entity is drawn animated, *all* instances of that entity must be drawn animated, or else they will not reset the pntbl appropriately.
    static MATRIX newMtx;
    slGetMatrix(newMtx);

	static FIXED m0x[4];
	static FIXED m1y[4];
	static FIXED m2z[4];
	
	m0x[0] = newMtx[X][X];
	m0x[1] = newMtx[Y][X];
	m0x[2] = newMtx[Z][X];
	m0x[3] = newMtx[3][X];
	
	m1y[0] = newMtx[X][Y];
	m1y[1] = newMtx[Y][Y];
	m1y[2] = newMtx[Z][Y];
	m1y[3] = newMtx[3][Y];
	
	m2z[0] = newMtx[X][Z];
	m2z[1] = newMtx[Y][Z];
	m2z[2] = newMtx[Z][Z];
	m2z[3] = newMtx[3][Z];

    XPDATA * model = ent->pol[0];
    Uint32	cnt;
	Uint32	nbPt	= model->nbPoint;
	Uint32	nbPol	= model->nbPolygon;
    if ( (transVerts[0]+nbPt) >= MAX_VERTS) return;
    if ( (transPolys[0]+nbPol) >= MAX_POLYGONS) return;
   // Uint32 PreClippingflag;
	
	FIXED transNormY;
	short colorBank;

	unsigned short comm_p_mode = 5328; //Should add color mode table? Color mode is applied here.
	unsigned short ctrl = 2;

 if(AnimArea[anims].startFrm != animCtrl->startFrm && AnimArea[anims].endFrm != animCtrl->endFrm) // Check to see if the animation matches.
 {
	AnimArea[anims].uniform = animCtrl->uniform;
	AnimArea[anims].currentFrm = animCtrl->startFrm<<3;
	AnimArea[anims].startFrm = animCtrl->startFrm;
	AnimArea[anims].endFrm = animCtrl->endFrm;
 }
//
	Uint8 localArate;
    /**Sets the animation data**/
///Variable interpolation set
if(AnimArea[anims].uniform == false){
localArate = animCtrl->arate[AnimArea[anims].currentKeyFrm];
} else 
{localArate = 2;
}
	AnimArea[anims].currentFrm += (localArate * framerate)>>1;

   AnimArea[anims].currentKeyFrm = (AnimArea[anims].currentFrm>>3);
    if (AnimArea[anims].currentKeyFrm >= AnimArea[anims].endFrm)    {
        AnimArea[anims].currentFrm -= (AnimArea[anims].endFrm - AnimArea[anims].startFrm)<<3;
        AnimArea[anims].currentKeyFrm = AnimArea[anims].currentFrm>>3;
     } else if(AnimArea[anims].currentKeyFrm < AnimArea[anims].startFrm){
		AnimArea[anims].currentKeyFrm = AnimArea[anims].startFrm;
		AnimArea[anims].currentFrm += (AnimArea[anims].endFrm-AnimArea[anims].startFrm)<<3;
	}
    Uint8 nextKeyFrm = AnimArea[anims].currentKeyFrm+1;
    if (nextKeyFrm >= AnimArea[anims].endFrm){
        nextKeyFrm = AnimArea[anims].startFrm;
	} else if (nextKeyFrm <= AnimArea[anims].startFrm){
        nextKeyFrm = AnimArea[anims].startFrm+1;
	}
    compVert * curKeyFrame = (compVert*)ent->animation[AnimArea[anims].currentKeyFrm]->cVert;
    compVert * nextKeyFrame = (compVert*)ent->animation[nextKeyFrm]->cVert;
	///Don't touch this! **absolute** frame delta 
	Uint32 frDelta = (AnimArea[anims].currentFrm)-(AnimArea[anims].currentKeyFrm<<3);

//Animation Data
    Sint32 * dst = model->pntbl[0]; //This pointer is incremented by the animation interpolator.
    short * src = curKeyFrame[0];
    short * nxt = nextKeyFrame[0];
 
    for (cnt=0; cnt<nbPt; cnt++)
    {
    /**Uncompress the vertices and apply linear interpolation**/
	animation_interpolator(src, nxt, frDelta, dst); //X
	animation_interpolator(src, nxt, frDelta, dst); //Y
	animation_interpolator(src, nxt, frDelta, dst); //Z
	//
        /**calculate z**/
        ssh2VertArea[cnt].pnt[Z] = trans_pt_by_component(&model->pntbl[cnt], m2z);
		
		ssh2VertArea[cnt].pnt[Z] = (ssh2VertArea[cnt].pnt[Z] > nearP) ? ssh2VertArea[cnt].pnt[Z] : nearP;

 
         /**Starts the division**/
        SetFixDiv(MsScreenDist, ssh2VertArea[cnt].pnt[Z]);

        /**Calculates X and Y while waiting for screenDist/z **/
        ssh2VertArea[cnt].pnt[Y] = trans_pt_by_component(&model->pntbl[cnt], m1y);
		ssh2VertArea[cnt].clipFlag = (ssh2VertArea[cnt].pnt[Z] > nearP) ? 0 : 16; //I have to put this here because my ASM is sloppy
        ssh2VertArea[cnt].pnt[X] = trans_pt_by_component(&model->pntbl[cnt], m0x);
		
        /** Retrieves the result of the division **/
		ssh2VertArea[cnt].inverseZ = *DVDNTL;

        /**Transform X and Y to screen space**/
        ssh2VertArea[cnt].pnt[X] = fxm(ssh2VertArea[cnt].pnt[X], ssh2VertArea[cnt].inverseZ)>>16;
        ssh2VertArea[cnt].pnt[Y] = fxm(ssh2VertArea[cnt].pnt[Y], ssh2VertArea[cnt].inverseZ)>>16;
 
        /**Clipping flags for fast culling/preclipping disable/enable and portal clipping **/
		ssh2VertArea[cnt].clipFlag |= (JO_ABS(ssh2VertArea[cnt].pnt[X]) > JO_TV_WIDTH_2) ? 1 : 0; //Simplified to increase CPU performance
		ssh2VertArea[cnt].clipFlag |= (JO_ABS(ssh2VertArea[cnt].pnt[Y]) > JO_TV_HEIGHT_2) ? 1 : 0; 
    }

    transVerts[0]+=nbPt;

    dst = (Sint32 *)&model->pltbl[0];
    Uint8 *src2 = ent->animation[AnimArea[anims].currentKeyFrm]->cNorm; //A new 1-byte src

    /**POLYGON PROCESSING**/ 
    for (cnt=0; cnt<nbPol; cnt++)
    {
		//Components of screen-space cross-product used for backface culling.
		//Vertice order hint:
		// 0 - 1
		// 3 - 2
		//A cross-product can tell us if it's facing the screen. If it is not, we do not want it.
		register int cross0 = (ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[X] - ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[X])
							* (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Y] - ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Y]);
		register int cross1 = (ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt[Y] - ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt[Y])
							* (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[X] - ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[X]);
		//Sorting target. Uses average of top-left and bottom-right. Adding logic to change sorting per-polygon HAMMERS performance in unacceptable ways.
		register int zDepthTgt = (ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt[Z] + ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt[Z])>>1;
 
 
		if(cross0 >= cross1 || zDepthTgt <= nearP || zDepthTgt >= farP ||
		ssh2VertArea[model->pltbl[cnt].Vertices[0]].clipFlag != 0 ||
		ssh2VertArea[model->pltbl[cnt].Vertices[2]].clipFlag != 0) continue;
		//Preclipping disable logic is removed. It is not worth the CPU performance cost.
		//New normals in from animation normal table
        *dst++=ANORMS[*src2][X];
        *dst++=ANORMS[*src2][Y];
        *dst++=ANORMS[*src2][Z];
		src2++;
        dst++; //Skip first two 16-bit vertice IDs
		dst++; //Skip last two 16-bit vertice IDs
		//
		//Transform the polygon's normal by light source vector
		transNormY = norm_component_trans(&model->pltbl[cnt].norm, lightSrc);
		//Use transformed normal as shade determinant
		colorBank = (transNormY < -32768) ? 0 : 1;
		colorBank = (transNormY > 16384) ? 2 : colorBank;
		colorBank = (transNormY > 32768) ? 3 : colorBank; 

	//Issues: Unless I load the data the same way in which XL2 does, I will have to adapt the PDATA ATTBL to these functions.
        ssh2SetCommand(ssh2VertArea[model->pltbl[cnt].Vertices[0]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[1]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[2]].pnt,
		ssh2VertArea[model->pltbl[cnt].Vertices[3]].pnt,
                     ctrl, comm_p_mode,
                     (int)__jo_sprite_pic[model->attbl[cnt].texno].data, colorBank<<6, __jo_sprite_def[model->attbl[cnt].texno].size, 0, zDepthTgt);
    }
		transPolys[0]+=nbPol;
		anims++;
}


